SystemOrganization addCategory: #LDAS!

Object subclass: #LDAllocator
	instanceVariableNames: 'topics topicCount featureCount documentCount topicPriors perDocumentTopics perDocumentFrequency perTopicFeatures perTopicFrequency perWordTopic featurePriors documents'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'LDAS'!

!LDAllocator commentStamp: 'toma 8/6/2015 01:08' prior: 0!
LDAllocator trains a Latent Dirichlet Allocation (LDA) Model on a given set of documents. It takes a set of documents (vectors of features) as input, clusters co-occuring features into (latent) topics and assigns each document an affinity towards each topic.

Documents need to be represented as sequences of natural numbers. Example:

	sixDocs := #(  #(1 2 3)  #(1 2 2 2)  #(1 1 3)  #(3 4 5) #(4 5) #(3 5 4 3) )

	(You can use LDFeatureSpace to map any hashable type to such lists)

We know we have 5 distinct features (1 to 5) and want to have them arranged in two topics:

	lda := LDAllocator new.
	model := lda allocate: sixDocs iterating: 10 features: 5topics: 2.

Let's inspect the topics:

	model featuresForTopicAt: 1
		#(4.0 4.0 2.0 1.0 0.0)
		which means features 1 and 2 are in that topic, 3 and 4 slightly.
	model featuresForTopicAt: 2
		 #(0.0 0.0 3.0 2.0 3.0)
		which means features 3, 4, and 5 reside in this topic.

Let's inspect which topics a document belongs to:

	model topicsForDocumentAt: 6
		#(0.0 4.0)
		which shows strong correspondence to topic 2 (containing 3, 4, and 5 as features).

Let's compute the similarity between two documents based on what we know about their topics:

	model documentSimilarityOf: 1 and: 2
		12.0
	model documentSimilarityOf: 5 and: 6
		4.0
	model documentSimilarityOf: 1 and: 6
		0.0

	which indicates that documents 1 and 2 are concerned with the same thing, 5 and 6 approximately with the same thing, and 1 and 6 are different concerns.
	



!

!LDAllocator methodsFor: 'as yet unclassified' stamp: 'toma 8/6/2015 00:29'!
allocate: docs iterating: numIterations features: numFeatures topics: numTopics
	"Train model with given documents for given number of iterations."

	self setupForDocuments: docs
			featureCount: numFeatures
			topicCount: numTopics;
		randomize.
	(1 to: numIterations) do: [ :i | self update].
	^self model
	
	! !

!LDAllocator methodsFor: 'as yet unclassified' stamp: 'toma 8/5/2015 16:51'!
associateFeature: feature withDoc: doc andTopic: topic
	
	self associateFeature: feature withDoc: doc andTopic: topic by: 1.! !

!LDAllocator methodsFor: 'as yet unclassified' stamp: 'toma 8/5/2015 18:36'!
associateFeature: feature withDoc: doc andTopic: topic by: delta
	"update statistics to reflect association between word, topic and document"

	perDocumentTopics at: doc at: topic put: 
		(delta + (perDocumentTopics at: doc at: topic)).

	perTopicFeatures at: topic at: feature put: 
		(delta + (perTopicFeatures at: topic at: feature)).

	perDocumentFrequency at: doc put:
		(delta + (perDocumentFrequency at: doc)).

	perTopicFrequency at: topic put:
		(delta + (perTopicFrequency at: topic)).! !

!LDAllocator methodsFor: 'as yet unclassified' stamp: 'toma 8/5/2015 16:51'!
dissociateFeature: feature withDoc: doc andTopic: topic
	
	self associateFeature: feature withDoc: doc andTopic: topic by: (1 negated).! !

!LDAllocator methodsFor: 'as yet unclassified' stamp: 'toma 8/5/2015 16:20'!
initialize
	! !

!LDAllocator methodsFor: 'as yet unclassified' stamp: 'toma 8/5/2015 18:55'!
model
	^(LDModel new)
			initializeTopics: perTopicFeatures 
			andDocuments: perDocumentTopics 
			priors: topicPriors;
		yourself.! !

!LDAllocator methodsFor: 'as yet unclassified' stamp: 'toma 8/5/2015 16:51'!
randomize
	"Initialize the topic model by assigning a random topic to each word"

	| topic  random |

	random := Random new.
	
	documents withIndexDo: [ :doc :di |
		doc withIndexDo: [ :word :wj |
			topic := random nextInt: topicCount.
			self associateFeature: word withDoc: di andTopic: topic. 	
			perWordTopic at: (di @ wj) put: topic.	
		]
	]! !

!LDAllocator methodsFor: 'as yet unclassified' stamp: 'toma 8/5/2015 16:32'!
setupForDocuments: docs featureCount: features topicCount: topics
	
	documents := docs.
	topicCount := topics.
	featureCount := features.
	documentCount := docs size.
	
	perDocumentTopics := Matrix
		rows: documentCount 
		columns: topicCount
		element: 0.0.
	perDocumentFrequency := (Array new: documentCount) atAllPut: 0.
	perTopicFeatures := Matrix
		rows: topicCount
		columns: featureCount
		element: 0.0.
	perTopicFrequency := (Array new: topicCount) atAllPut: 0.

	topicPriors := (Array new: topicCount) atAllPut: (1.0 / topicCount).
	featurePriors := (Array new: featureCount) atAllPut: (1.0 / featureCount).

	perWordTopic := Dictionary new.

	


	
	! !

!LDAllocator methodsFor: 'as yet unclassified' stamp: 'toma 8/5/2015 18:08'!
topicProbabilityOver: doc andFeature: feature
	"Compute P(t | d, f) given previous topic allocation and priors.
	The result is a distribution, not a value!!"

	| featurePrior topics docFrequency |

	featurePrior := featurePriors at: feature.
	topics := perDocumentTopics atRow: doc.
	docFrequency := perDocumentFrequency at: doc.

	^ LDMultinomial normalized: (
		(1 to: topicCount) collect: [ :topic | | pFeature pTopic |

		"bayesian estimates proportional to feature and topic marginals
		(with normalization terms, don't ask!!)"

		pFeature := (perTopicFeatures at: topic at: feature) + featurePrior.		
		pFeature := pFeature / (1.0 + (perTopicFrequency at: topic)).

		pTopic := (topics at: topic) + (topicPriors at: topic).
		pTopic := pTopic / (docFrequency + ((topicPriors at: topic) * topicCount)).

		pFeature * pTopic		
	])! !

!LDAllocator methodsFor: 'as yet unclassified' stamp: 'toma 8/5/2015 18:16'!
update
	| topic | 

	documents withIndexDo: [:document :di |
		document withIndexDo: [:word :wj |
			topic := perWordTopic at: (di @ wj).
			self dissociateFeature: word withDoc: di andTopic: topic.
			topic := (self topicProbabilityOver: di 
 							andFeature: word) sample.
			self associateFeature: word withDoc: di andTopic: topic.
			perWordTopic at: (di @ wj) put: topic.
		]
	]! !

Object subclass: #LDFeatureSpace
	instanceVariableNames: 'featureIdMap idFeatureMap maxId'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'LDAS'!

!LDFeatureSpace commentStamp: 'toma 8/6/2015 00:51' prior: 0!
This class helps to transform arbitrary hashable features (e.g. words or collections of words) into natural numbers. The transformation is required, as LDAllocator only accepts integer features.

Example:

	space := LDFeatureSpace new.
	space representAll: #(a b c 100).
		gives #(1 2 3 4)
	space representAll: #(a c d).
		subsequently gives  #(1 3 5) because a and c are already known as feature #1 and #3, c is inserted as new feature #5.

The space needs to be retained for later interpretation of encoded results:

	space interpret: 2
		gives #b
	space interpretAll: #(5 3 2 1)
		gives #(#d #c #b #a)
!

!LDFeatureSpace methodsFor: 'as yet unclassified' stamp: 'toma 8/6/2015 00:34'!
featureCount
	
	^maxId! !

!LDFeatureSpace methodsFor: 'as yet unclassified' stamp: 'toma 8/6/2015 00:33'!
initialize 
	featureIdMap := Dictionary new.
	idFeatureMap := Dictionary new.
	maxId := 0.! !

!LDFeatureSpace methodsFor: 'as yet unclassified' stamp: 'toma 8/6/2015 00:39'!
interpret: aFeatureId
	
	^idFeatureMap at: aFeatureId! !

!LDFeatureSpace methodsFor: 'as yet unclassified' stamp: 'toma 8/6/2015 00:41'!
interpretAll: aCollection
	
	^aCollection collect: [ :each | self interpret: each ]! !

!LDFeatureSpace methodsFor: 'as yet unclassified' stamp: 'toma 8/6/2015 00:39'!
represent: aFeature
	
	^featureIdMap at: aFeature ifAbsentPut: [ 
		maxId := maxId + 1.
		idFeatureMap at: maxId put: aFeature.
		maxId]! !

!LDFeatureSpace methodsFor: 'as yet unclassified' stamp: 'toma 8/6/2015 00:40'!
representAll: aCollection

	^aCollection collect: [ :each | self represent: each ]! !

Object subclass: #LDModel
	instanceVariableNames: 'perTopicFeatures perDocumentTopics topicPriors'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'LDAS'!

!LDModel commentStamp: 'toma 8/6/2015 01:08' prior: 0!
This is the result of LDAllocator training. See LDAllocator documentation for examples.!

!LDModel methodsFor: 'as yet unclassified' stamp: 'toma 8/5/2015 18:52'!
documentSimilarityOf: doc1 and: doc2
	
	^( (self topicsForDocumentAt: doc1)
		with: (self topicsForDocumentAt: doc2) collect: [ :i :k | i * k] ) sum! !

!LDModel methodsFor: 'as yet unclassified' stamp: 'toma 8/5/2015 18:48'!
documentsForTopicAt: topic
	
	^perDocumentTopics atColumn: topic! !

!LDModel methodsFor: 'as yet unclassified' stamp: 'toma 8/5/2015 18:53'!
featureSimilarityOf: word1 and: word2
	
	^( (self topicsForFeatureAt: word1)
		with: (self topicsForFeatureAt: word2) collect: [ :i :k | i * k] ) sum! !

!LDModel methodsFor: 'as yet unclassified' stamp: 'toma 8/5/2015 18:45'!
featuresForTopicAt: topic

	^perTopicFeatures atRow: topic! !

!LDModel methodsFor: 'as yet unclassified' stamp: 'toma 8/5/2015 18:57'!
initializeTopics: topics andDocuments: documents priors: priors
	
	perTopicFeatures := topics.
	perDocumentTopics := documents.
	topicPriors := priors.! !

!LDModel methodsFor: 'as yet unclassified' stamp: 'toma 8/5/2015 18:45'!
topicsForDocumentAt: docid
	
	^perDocumentTopics atRow: docid! !

!LDModel methodsFor: 'as yet unclassified' stamp: 'toma 8/5/2015 18:46'!
topicsForFeatureAt: word

	^perTopicFeatures atColumn: word! !

Object subclass: #LDMultinomial
	instanceVariableNames: 'probabilities'
	classVariableNames: 'Rand'
	poolDictionaries: ''
	category: 'LDAS'!

!LDMultinomial commentStamp: 'toma 8/6/2015 01:09' prior: 0!
This represents a multinomially distributed random variable and is used internally by LDAllocator to deal with multinomial distributions.!

!LDMultinomial class methodsFor: 'as yet unclassified' stamp: 'toma 8/5/2015 17:58'!
initialize
	"LDMultinomial initialize"

	Rand := Random new.! !

!LDMultinomial class methodsFor: 'as yet unclassified' stamp: 'toma 8/5/2015 18:37'!
normalized: observation
	^ (LDMultinomial new)
		observe: observation;
		yourself.! !

!LDMultinomial methodsFor: 'as yet unclassified' stamp: 'toma 8/5/2015 17:55'!
normalize
	| sum |
	
	sum := probabilities sum.
	probabilities := probabilities collect: [:p | p / sum]! !

!LDMultinomial methodsFor: 'as yet unclassified' stamp: 'toma 8/5/2015 17:56'!
observe: frequencies
	
	probabilities := frequencies.
	self normalize! !

!LDMultinomial methodsFor: 'as yet unclassified' stamp: 'toma 8/5/2015 18:04'!
sample
	"Draw a sample index from the distribution"
	| r |
	
	r := Rand next.
	probabilities withIndexDo: [ :p :i |
		r := r - p.
		(r < 0) ifTrue: [^ i]]! !
LDMultinomial initialize!
